/* This is the template file for HW3_5. You need to complete the code by 
 * implementing the parts marked by FIXME. 
 *
 * Submit your completed code to the TA by email. Also submit the plots 
 * and your conclusions of the comparative study.
 *
 * To compile, use the UNIX command 
 *      make
 * in the source directory, and it would then invoke the compilation command 
 * with the included makefile. You may sometimes need to use the command 
 *      make clean
 * before recompiling.
 *
 * This will generate an executable hw3_5. The -g option is optional and 
 * is needed only if you need to debug the program in a debugger such as ddd.
 * The -Wall option would enable compiler's warning messages.
 *
 * Run the program with command
 *      ../hw3_5
 * which would generate a M-file, which you can use to generate the plots by
 * issueing the UNIX command 
 *      make plot
 * in the source directory. It will then generate two PDF files, which you
 * should submit to the TA along with your source code and your conclusions.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

/* Define types for matrix and vector. 
 * Access the (i,j)-entries of a matrix A using A.vals[i][j], and 
 * access the ith entries of a vector x using x.vals[i], where 
 * i and j start from 0.
 */

typedef struct {
    double** vals;
    int m;
    int n;
} Matrix;

typedef struct {
    double*  vals;
    int m;
} Vector;

/* Prototypes for local functions. 
   You need to implement some of these functions. */
void  lsqr_Householder( Vector x, const Matrix A, const Vector b);
void  householder( Matrix V, Matrix A);
void  implicitQtb( Vector Qtb, const Matrix V, const Vector b);

void  lsqr_mgs( Vector x, const Matrix A, const Vector b);
void  mgs(Matrix Q, Matrix R, const Matrix A);

void  backsolve( Vector x, const Matrix A, const Vector b);
double norm2_col( double *vec, int vec_dim);

/* Define prototypes of helper functions */
Matrix allocate_matrix(int m, int n);
Vector allocate_vector( int m);
void   deallocate_matrix( Matrix);
void   deallocate_vector( Vector);

void   write_vec(FILE *fid, char *name, const double *vec, int n);
double tictoc (int n);

#define MAXDEGREE 15
#define MAXPOINTS 30

int main(int argc, char **argv) 
{   
    double ts[MAXPOINTS];
    double dtemp[MAXDEGREE];
    double err_house[MAXDEGREE-2], err_mgs[MAXDEGREE-2];
    double times_house[MAXDEGREE-2], times_mgs[MAXDEGREE-2];

    FILE *fid;
    int m, n, i, j;
    int niter = 10000;

    /* Set n to degree of polynomial */
    for (n=3; n<=MAXDEGREE; ++n) {
        Matrix A; 
        Vector b,x;

        /* Set m to the number of points */
        m = n + n;

        A = allocate_matrix( m, n);
        b = allocate_vector( m);
        x = allocate_vector( n);

        /* Compute Vandermonde matrix A and right-hand side b */
        for (i=0; i<m; i++) {
            ts[i] = (((double)i)) / (((double)(m - 1)));

            A.vals[i][0] = 1.0;
            b.vals[i] = 1.0;
        }

        for (j=1; j<n; j++) {
            for (i=0; i<m; i++) {
                A.vals[i][j] = A.vals[i][j-1] * ts[i];
                b.vals[i] = 1.0 +  b.vals[i] * ts[i];
            }
        }

        tictoc(1);
        /* Invoke least-squares solver with Householder QR with implicit Q */
        /* Run for many iterations to obtain more accurate timing results */
        for (j=0; j<niter; ++j) {
            lsqr_Householder(x, A, b);
        }
        times_house[n - 3] =  tictoc(2) / niter;

        for (j=0; j<n; j++) {
            dtemp[j] =  x.vals[j] - 1.0;
        }
        err_house[n - 3] = norm2_col(dtemp, n);

        tictoc(1);
        /* Invoke least-squares solver with Modified Gram-Schmidt */
        /* Run for many iterations to obtain more accurate timing results */
        for (j=0; j<niter; ++j) {
            lsqr_mgs(x, A, b);
        }
        times_mgs[n - 3] =  tictoc(2) / niter;

        for (j=0; j<n; j++) {
            dtemp[j] =  x.vals[j] - 1.0;
        }
        err_mgs[n - 3] = norm2_col(dtemp, n);

        deallocate_matrix( A);
        deallocate_vector( b);
        deallocate_vector( x);
    }

    /* Write out results into a Matlab file */
    fid = fopen("results.m", "w");
    write_vec(fid, "err_house", err_house, MAXDEGREE-2);
    write_vec(fid, "err_mgs  ", err_mgs, MAXDEGREE-2);
    write_vec(fid, "times_house", times_house, MAXDEGREE-2);
    write_vec(fid, "times_mgs", times_mgs, MAXDEGREE-2);
    fclose(fid);

    /* plotresults; */
    return 0;
}

/*************************************************************
 *
 * FUNCTION: mgs
 *
 * Modified Gram-Schdmit Algorithm
 *************************************************************/

void  mgs(Matrix Q, Matrix R, const Matrix A) {
    int i,j,k;

    /* Q = A; */
    for (i = 0; i < A.m; ++i)
        for (j = 0; j < A.n; ++j)
            Q.vals[i][j] = A.vals[i][j];

    for (i=0; i<A.n; ++i) {
        /* R(i,i) = norm (Q(:,i), 2); */
        R.vals[i][i] = 0.0;
        for (k=0; k<A.m; ++k)
            R.vals[i][i] += Q.vals[k][i] * Q.vals[k][i];
        R.vals[i][i] = sqrt(R.vals[i][i]);

        /* Q(:,i) = Q(:,i) ./ R(i,i); */
        for (k = 0; k < A.m; ++k)
            Q.vals[k][i] /= R.vals[i][i];
		
        for (j=i+1; j<A.n; ++j) {
            /* R(i,j) = Q(:,i)' * Q(:,j); */
            R.vals[i][j] = 0.0;
            for (k=0; k<A.m; ++k)
                R.vals[i][j] += Q.vals[k][i] * Q.vals[k][j];

            /* Q(:,j) = Q(:,j) - R(i,j) * Q(:,i); */
            for (k=0; k<A.m; ++k)
                Q.vals[k][j] -= R.vals[i][j] * Q.vals[k][i];
        }
    }
}

/*************************************************************
 *
 * FUNCTION: lsqr_Householder
 *
 * Least squares via Householder QR factorization.
 *************************************************************/

void  lsqr_Householder(
   Vector x, 
   const Matrix A, 
   const Vector b)
{
    Matrix V, R;
    Vector Qtb;
    int i, j;

    V = allocate_matrix( A.m, A.n);
    R = allocate_matrix( A.m, A.n);
    Qtb = allocate_vector( b.m);

    for (i=0; i<A.m; ++i)
        for (j=0; j<A.n; ++j)
            R.vals[i][j] = A.vals[i][j];

    /* Perform Householder QR factorization with implicit Q matrix */
    householder(V, R);

    /* Compute implicit vector multiplication  */
    implicitQtb(Qtb, V, b);

    /* Solve the upper-triangular system Rx=Qtb */
    backsolve(x, R, Qtb);

    deallocate_matrix(V);
    deallocate_matrix(R);
    deallocate_vector(Qtb);
    return;
}

/*************************************************************
 *
 * FUNCTION: implicitQtb
 *
 * Implicit calculation of product Q'*b, where Q is stored as 
 *     reflection vectors.
 *************************************************************/

void  implicitQtb(
    Vector Qtb,
    const Matrix V,
    const Vector b)
{
    int i, k;

    /* Copy b to Qtb */
    for (k=0; k<V.m; k++)
        Qtb.vals[k] = b.vals[k];
   
    for (k=0; k<V.n; k++) {
        double s = 0.0;
        /* Compute v_k * b(k:m) */
        for (i=k; i< V.m; i++)
            s += ( V.vals[i][k] *  Qtb.vals[i]);

        /* Compute b(k:m) = b(k:m) - 2*s*v_k. */
        /* FIXME */
    }
}

/*************************************************************
 *
 * FUNCTION: lsqr_mgs
 *
 * Least squares via Modified Gram-Schmidt orthogonalization.
 *************************************************************/

void  lsqr_mgs(
    Vector x, 
    const Matrix A, 
    const Vector b)
{
    Matrix Q, R;
    Vector Qtb;
    int i,j;

    Q = allocate_matrix( A.m, A.n);
    R = allocate_matrix( A.n, A.n);
    Qtb = allocate_vector( b.m);

    /* Perform Gram-Schmidt orthogonalization */
    mgs( Q, R, A);

    /* Compute vector multiplication Q'*b  */
    /* FIXME */

    /* Solve the upper-triangular system Rx=Qtb */
    backsolve(x, R, Qtb);

    deallocate_matrix(Q);
    deallocate_matrix(R);
    deallocate_vector(Qtb);
}

/*************************************************************
 *
 * FUNCTION: backsolve
 *
 * Perform back substitution.
 *************************************************************/

void  backsolve( 
    Vector x,
    const Matrix A, 
    const Vector b)
{
    int i, j;
    int n = A.n;

    x.vals[n - 1] =  b.vals[n - 1] /  A.vals[n-1][n-1];

    for (i=n-2; i>=0; i--) {
        /* Implement this part of the code yourself. */

        double dtemp = 0.0;
        for (j=i+1; j<n; j++) {
            dtemp = dtemp +  A.vals[i][j] *  x.vals[j];
        }

        x.vals[i] = ( b.vals[i] - dtemp) /  A.vals[i][i];
    }
    return;
}

/*************************************************************
 *
 * FUNCTION: norm2_col
 *
 * Compute 2-norm of a given vector.
 * Note: It is not a perfect implementation as it does not guard against 
 *       overflow or underflow.
 *************************************************************/

double norm2_col(
    double *vec, 
    int vec_dim)
{
    double v_out;
    int i;

    v_out = 0.0;
    for (i=0; i<vec_dim; i++) {
        v_out = v_out +  vec[i] *  vec[i];
    }
    return sqrt(v_out);
}

/*************************************************************
 *
 * FUNCTION: householder
 *
 * Householder QR factorization.
 * A is the input matrix, but its upper-triangular part stores R at output.
 * V is an output, but it should have already been allocated at input.
 * At output, the lower-triangular part of V (including the diagonals) 
 *   stores the Householder reflection vectors v_k.
 * 
 * Note: This is not a memory efficient implementation, because V could have 
 *   been stored in the lower-triangular part of A along with a vevctor.
 *************************************************************/

void  householder(
   Matrix V,
   Matrix A)
{
    static double x[MAXPOINTS];
    int x_dim, i, j, k;
    double dtemp;

    for (k=0; k<A.n; k++) {
        /* Compute x. */
        x_dim = A.m - k;

        for (i=0; i< x_dim; i++) 
            x[i] = A.vals[i+k][k];

        /* Compute v_k and store in x */
        if ( x[0] >= 0.0)
            x[0] += norm2_col(x, x_dim);
        else
            x[0] -= norm2_col(x, x_dim);

        dtemp = norm2_col(x, x_dim);
        for (i=0; i<x_dim; i++) 
            x[i] /= dtemp;

        /* Update matrix A to compute R */
        /* A(k:m, k:n) = A(k:m, k:n) - v_k * (2 * v_k' * A(k:m, jj)); */
        /* FIXME */

        /* Store v_k into V */
        for (i=k; i<A.m; i++)
            V.vals[i][k] = x[i-k];
    }
}

/*-----------------------------------------------------------------------------
 * FUNCTION: allocate_matrix - Allocate memory for a given matrix.
 * --------------------------------------------------------------------------*/
/* Disclaimer: The approach used here is not the most efficient way for
 * implementing matrices, but it is adopted here for convenience.
 */
Matrix allocate_matrix(int m, int n) {
    Matrix A;
    double *ptmp;
    int i;

    A.m = m; A.n = n;
    ptmp = (double *)malloc( sizeof(double) * m * n);
    A.vals = (double **)malloc( sizeof(double **) * m);

    for (i=0; i<m; ++i) {
        A.vals[i] = ptmp + i*n;
    }

    return A;
}

/*-----------------------------------------------------------------------------
 * FUNCTION: deallocate_matrix - De-allocate memory for a matrix.
 * --------------------------------------------------------------------------*/
void deallocate_matrix( Matrix A) {
    free( A.vals[0]);
    free( A.vals); A.vals=NULL;
}


/*-----------------------------------------------------------------------------
 * FUNCTION: allocate_vector - Allocate memory for an m vector.
 * --------------------------------------------------------------------------*/
Vector allocate_vector( int m) {
    Vector x;

    x.m = m;
    x.vals = (double *)malloc( sizeof(double) * m);
    return x;
}

/*-----------------------------------------------------------------------------
 * FUNCTION: deallocate_vector - De-allocate the memory for an m vector. 
 * --------------------------------------------------------------------------*/
void deallocate_vector( Vector x) {
    free( x.vals); x.vals=NULL;
}

#include <time.h>
#include <sys/time.h>

/*-----------------------------------------------------------------------------
 * FUNCTION: tictoc - Initialize (1) or get (2) the elapsed time since in seconds
 * --------------------------------------------------------------------------*/
double tictoc (int n)
{
    double    y = 0.0;
    static    struct timeval start_time, end_time; 

    if (n == 1) { 
        gettimeofday(&start_time, NULL);
    }
    else if (n == 2) { 
        gettimeofday(&end_time, NULL);

        y = (end_time.tv_sec - start_time.tv_sec) + (end_time.tv_usec-start_time.tv_usec)*1.e-6;
    }
    return (y); 
}

/*************************************************************
 *
 * FUNCTION: write_vec
 *
 * Write out a vector into file
 *************************************************************/

void  write_vec(FILE *fid, char *name, const double *vec, int n) {
    int i;

    fprintf(fid, "%s=[", name);
    for (i=0; i<n; ++i) {
        fprintf(fid, "%g; ", vec[i]);
    }
    fprintf(fid, "];\n");
}
